/********************************************************************
*				Program agnus.cpp						  		 	*
*		Primjena GA na problem rjesavanja problema	prekrivanja		*
* 		povrsine. 													*
*		Ulaz programa je: prva linija X Y L	  						*
*		gdje su X*Y dimenzije polja, a L broj svjetiljki.			*
*		Nakon toga slijedi L linija oblika: xi yi, gdje su 			*
* 		xi i yi koordinate unutar polja 1<=xi<=X, 1<=yi<=Y		    *
*																	*
*		Parametri programa su:										*
*			mutcoef	- double - dinamika mijenjanja mutacije			*
*						1.0 = konstantna mutacija					*
*			mutation - double - koeficijent mutacije				*
*			select 	- double - koeficijent mutacije					*
*			pop	- int - velicina populacije							*
* 			breeding - {0,1} - izbor krizanja, 1=uniformno			*
*											   0=tockom prekida		*
*			A, B - int - parametri funkcije kazne					*
*			verbose - bez vrijednosti, ukljucuje ispis			   	*										   
*																	*
*		Primjer pokretanja programa:								*
* 		neka se u datoteci ulaz.txt nalaze slijedece linije			*
*		5 6 4														*
*		1 1															*
*		2 4															*
* 		3 1															*
* 		5 3															*
*																	*
*		Program mozemo pokrenuti kao:								*
*		agnus.exe mutcoef=1.0 mutation=0.003 select=0.5 pop=100 	*
*					breeding=1 A=1 B=2 verbose < ulaz.txt 			*
*																	*
*																	*
*		Ako neki od parametara izostavimo, on poprima vrijednost	*
*		defaultne postavke											*
*																	*
*********************************************************************/

#include <iostream>
#include <cstdio>
#include <cstring>
#include <vector>
#include <set>
#include <algorithm>
#include <iterator>
using namespace std;


#define K 1
#define MAXINT 1000000007

int R,C,L;
vector < pair <int,int> > lamp_pos;
vector <char> mora;
int BEST_SOLUTION=MAXINT;
int NAJMANJA_KAZNA=MAXINT;

/* DEFAULTNE POSTAVKE */
int POPULATION = 102;
int SELECT=POPULATION/2;
int RUNS=5000;
int A=1;
int B=2;
double MUTATION = 0.003;
int same=1000;
double mutcoef=1.0;
bool breeding=1;
bool verbose=false;
/***********************/


// Funkcija provjerava je li par tocaka unutar polja
bool inside( pair <int,int> pos) {
 	if (pos.first>=0 and pos.first<R and pos.second>=0 and pos.second<=C)
 		return true;
	else
		return false;
}    


// Klasa predstavlja jednu jedinku populacije
class state {
	public:	
	vector<char> lampa;
	int nepokriveno;
   	int broj;
   	int kazna;	
   
	state() {
		nepokriveno=broj=kazna=0;
	}	
	state(vector <char> a) {
		lampa=a;
		nepokriveno=broj=kazna=0;
	}
	bool operator < (const state &a) const {
		/*if (nepokriveno!=0 and a.nepokriveno==0) return true;
		if (nepokriveno==0 and a.nepokriveno==0) return broj<a.broj;
		return nepokriveno<a.nepokriveno;	*/
		return lampa < a.lampa;
	}
	bool operator == (const state &a) const {
   		for (int i=0;i<lampa.size();i++)
     		if (lampa[i]!=a.lampa[i]) return false;
       	return true;  
  	}     
	void add(const vector<char> &a) {
		for (int i=0;i<lampa.size();i++)
			lampa[i] = lampa[i] or a[i];
	}
	void eval(void) {
	    broj=0;
     	vector < vector<char> > polje(R);
		for (int i=0;i<R;i++) 
			polje[i].resize(C);
		for (int i=0;i<lampa.size();i++) {
			if (lampa[i]==true) {
				broj++;
				int x=lamp_pos[i].first,y=lamp_pos[i].second;
				for (int q=x-1;q<=x+1;q++)
					for (int w=y-1;w<=y+1;w++)
						if (inside(make_pair(q,w)))
							polje[q][w]=true;
			}
		}

		for (int x=0;x<R;x++)
			for (int y=0;y<C;y++) 
				if (polje[x][y]==false) 
					nepokriveno++;	    
	}
 	void mutate(void) {
    	for (int i=0;i<lampa.size();i++)
     		if (mora[i]==0 and rand()%1000<=1000*MUTATION) {
           		lampa[i]=lampa[i] xor 1;
          		nepokriveno=broj=kazna=0;  
         	}  		
   	}     
};    

vector <state> merge(const vector <state> &a,const vector <state> &b) {
	vector <state> result;
	result.reserve(a.size()+b.size());
	for (int i=0;i<b.size();i++)
		result.push_back(b[i]);
	return result;
}

/* GENERATE POPULATION */

vector <state> generate_population(int pop_size,int state_size) {
	int how=1;

	vector <state> pop;
	
	for (int i=0;i<pop_size;i++) {
		state adam;
		adam.lampa.resize(state_size);		
		for (int j=0;j<state_size;j++)
			if ((rand()%100)<51) adam.lampa[j]=0; else adam.lampa[j]=1;
		adam.add(mora);
		pop.push_back(adam);
	}
	
	return pop;	
}

/* PRINT POPULATION */
void print_population( vector <state> & pop) {
    printf ("\n	**	population print	**	");
	printf ("\n");
	for (int i=0;i<pop.size();i++) {
		for (int j=0;j<L;j++) {
			printf ("%d",pop[i].lampa[j]);
		}
		printf ("\t");
		printf ("broj: %d, nepokriveno: %d -> kazna %d\n",
				pop[i].broj,pop[i].nepokriveno,pop[i].kazna);
		
	}
}


// Operator mutacije
void mutation(vector <state> &pop) {
 	for (int i=0;i<pop.size();i++) 
     	pop[i].mutate();  
}    

    
// Funkcija kazne
void evaluate(vector <state> &pop) {
	for (int i=0;i<pop.size();i++) {
		if (pop[i].kazna==0) pop[i].eval();	
		int k = A*pop[i].broj + B*pop[i].nepokriveno; 
		pop[i].kazna = k;
		if (pop[i].nepokriveno==0 and pop[i].broj<BEST_SOLUTION) BEST_SOLUTION=pop[i].broj;
		if (k<NAJMANJA_KAZNA) NAJMANJA_KAZNA=k;
	}		
}


// Funkcija eliminacije

int selection(vector <state> &pop) {
	vector <int> kazna(pop.size());
	for (int i=0;i<pop.size();i++) kazna[i]=pop[i].kazna;

		
 	int mink=MAXINT,ukupnok=0;
 	int N=kazna.size();
 	vector <char> sel(N);
 	vector <int> help,index;
 	for (int i=0;i<N;i++)
 		if ( kazna[i] < mink ) mink=kazna[i];
	for (int i=0;i<N;i++)
		kazna[i]-=mink-K;
	
  	for (int i=0;i<N;i++) {
  	    if (kazna[i]!=0) {
	  		help.push_back(ukupnok);
	  		index.push_back(i);
    		ukupnok+=kazna[i]; 	   
  		} 		
  	}  
   	help.push_back(ukupnok);
    index.push_back(-1);
    
    for (int i=0;i<SELECT;i++) {
    	int done=false;
     	for (int j=0;j<20 and done==false;j++) {
        	int r=rand()%ukupnok;
         	int pos=upper_bound(help.begin(),help.end(),r)-help.begin()-1;
          	if (pos<0) perror("function selection(...): pos < 0\n");
          	if (sel[index[pos]]==0) {
          	    if (index[pos]<0 or index[pos]>=kazna.size()) 
          	    	perror("function selection(...): index not ok\n");
          	 	sel[index[pos]]=1;
             	done=true;   
          	}    
      	} 
  	}      

	vector <state> npop;
	for (int i=0;i<pop.size();i++)
		if (sel[i]==false) npop.push_back(pop[i]);
	pop=npop;
   	return 0;		
}    


// Uniformno krizanje
state uniform_child(state tata, state mama) {
	state beba;
	beba.lampa.resize(L);
	for (int i=0;i<L;i++) 
		if ( rand()%100 < 51 ) beba.lampa[i]=tata.lampa[i]; else beba.lampa[i]=mama.lampa[i];
	beba.add(mora);
	return beba;		
}

// Krizanje s jednom tockom prekida
state interrupt_child(state tata,state mama) {
    state beba;
    beba.lampa.resize(L);
 	int pos=rand()%L;
  	int tko=rand()%2;
   	for (int i=0;i<L;i++)
   		if (tko==1) 
	    	if (i<=pos)	beba.lampa[i]=tata.lampa[i]; else beba.lampa[i]=mama.lampa[i];   
	    else
     		if (i>pos)	beba.lampa[i]=tata.lampa[i]; else beba.lampa[i]=mama.lampa[i];
    beba.add(mora);		
	return beba;
}    

// Operator krizanja za cijelu populaciju
int cross( vector <state> &pop,int how_many) {
	vector <state> npop = pop;
	int N=pop.size();
	for (int i=0;i<how_many;i++) {
		int q = rand()%N;
		int w;
		while ( q!= (w=rand()%N ));

		if (pop[q]==pop[w]) pop[w].mutate();
		if ( breeding==1) 
			pop.push_back( uniform_child(pop[q],pop[w]) );
		else
			pop.push_back( interrupt_child(pop[q],pop[w]) );
	}
	return 0;
}

// Funkcija koja izracunava koje svjetiljke moraju biti upaljene
vector <char> izracunaj_mora( vector < pair<int,int> > lamp ) {
    vector < vector <int> > polje(R);
    mora.resize(L);
   	for (int i=0;i<R;i++)
    	polje[i].resize(L);

	for (int i=0;i<L;i++) {
		int x=lamp_pos[i].first;
  		int y=lamp_pos[i].second;
    
       	for (int j=x-1;j<=x+1;j++)
        	for (int k=y-1;k<=y+1;k++)
         		if ( inside(make_pair(j,k)) )
           		polje[j][k]++;   
	}   
	/* pogledaj postoji li neosvijetljeni dio i koji dijelovi
 		su samo jednom osvijetljeni */
   	for (int x=0;x<R;x++)
   		for (int y=0;y<C;y++) {
   		 	if (polje[x][y]==0) {
           		printf ("Polje se ne moze osvijetliti\n");
           		exit(0);
          	}       
          	if (polje[x][y]==1) {
          	 	for (int q=x-1;q<=x+1;q++)
             		for (int w=y-1;w<=y+1;w++) {
             		 	pair <int,int> P = make_pair(q,w);
                 		if (inside(P) and find(lamp_pos.begin(),lamp_pos.end(),P)!=lamp_pos.end())
                   			mora[find(lamp_pos.begin(),lamp_pos.end(),P)-lamp_pos.begin()]=true;	   
             		}    
          	}    
   		}    
	return mora;    
}


// Funkcija koja izracunava koliko je razlicitih jedinki u populaciji
int how_many_same( vector <state> &pop) {
    set <state> S;
    for (int i=0;i<pop.size();i++)
    	S.insert(pop[i]);
	if (verbose) printf (" %d razlicitih od %d\n",S.size(),POPULATION);
	int N=pop.size();
	vector <state> npop;
	set <state>::iterator SIT;
	for (SIT=S.begin();SIT!=S.end();SIT++)
		npop.push_back(*SIT);
	cross(npop,N-npop.size());
	pop=npop;
	return 0;
}

// Obrada ulaznih parametara
int parse_input(int argc,char *argv[]) {
	char var[20]={0};
	double val;
	for (int i=1;i<argc;i++) {
		for (int j=0;j<strlen(argv[i]);j++)
			if (argv[i][j]=='=') argv[i][j]=' ';

		//fprintf (stderr,"*%s*\n",argv[i]);
		sscanf(argv[i],"%s %lf\n",var,&val);
		
		if ( strcmp(var,"pop")==0 ) {
			POPULATION=(int)val;
		} else if ( strcmp(var,"select")==0 ) {
			SELECT=(int)(1.0*POPULATION*val);
		} else if ( strcmp(var,"mutcoef")==0 ) {
			mutcoef=val;
		} else if ( strcmp(var,"mutation")==0 ) {
			MUTATION=val;
		} else if ( strcmp(var,"A")==0 ) {
			A=(int)val;
		} else if ( strcmp(var,"B")==0 ) {
			B=(int)val;
		} else if ( strcmp(var,"runs")==0 ) {
			RUNS=(int)val;	
		} else if ( strcmp(var,"breeding")==0) {
			breeding=val;
		} else if ( strcmp(var,"verbose")==0 ) {
			verbose=true;
		} 	
				
		//fprintf (stderr,"%s = %lf\n",var,val);
	}
	return 0;
}

int main(int argc,char *argv[]) {

	parse_input(argc,argv);
	printf ("pop %d, select %lf, mutcoef %lf, A %d, B %d, mutation %lf, runs %d ",
			POPULATION,1.*SELECT/POPULATION,mutcoef,A,B,MUTATION,RUNS);
    
	srand(time(0));
    cin >> R >> C >> L;

    for (int i=0;i<L;i++) {
    	int a,b;
     	cin >> a >> b;
		a--;
		b--;
      	lamp_pos.push_back( make_pair(a,b)); 	   
    } 
	
	mora = izracunaj_mora(lamp_pos);
		
	
	vector <state> pop = generate_population(POPULATION,L);
	

	int same_count=1;
	for (int run=0;run<RUNS;run++) {
	    int temp_best=BEST_SOLUTION;
  		evaluate(pop);

		/* Ako je verbose mode ukljucen, svakih 100 generacija se ispisuje
			najbolje rjesenje i broj razlicitih rjesenja unutar populacije */
  		if (run%100==0) {
        	if (verbose==true) printf ("Generacija %d -> rjesenje %d (mutation:%lf)\n",run,BEST_SOLUTION,MUTATION);
    		how_many_same(pop);
     	}   	

		
    	selection(pop);
   		

       	cross(pop,SELECT);

        

        mutation(pop);		

   		if (temp_best==BEST_SOLUTION) {
         	same_count++; 
         	if (same_count%same==0) MUTATION*=mutcoef;
        } else 
        	same_count=1;
	}

	printf ("--> %d\n",BEST_SOLUTION);	
  	
	return 0;   
}    
